  --
  -- AP900
  -- Class for all sprayers
  --
  -- @author  Stefan Geiger
  -- @date  24/02/08
  --
  -- Copyright (C) GIANTS Software GmbH, Confidential, All Rights Reserved.
  
AP900 = {};
  
  
AP900.SPRAYTYPE_UNKNOWN = 0;
AP900.NUM_SPRAYTYPES = 0;
  
AP900.sprayTypes = {};
AP900.sprayTypeIndexToDesc = {};
  
AP900.sprayTypeToFillType = {};
AP900.fillTypeToSprayType = {};
  
function AP900.registerSprayType(name, pricePerLiter, litersPerSqmPerSecond, hudOverlayFilename)
      local key = "SPRAYTYPE_"..string.upper(name);
      if AP900[key] == nil then
          AP900.NUM_SPRAYTYPES = AP900.NUM_SPRAYTYPES+1;
          AP900[key] = AP900.NUM_SPRAYTYPES;
  
          local desc = {name = name, index = AP900.NUM_SPRAYTYPES};
          desc.pricePerLiter = pricePerLiter;
          desc.litersPerSqmPerSecond = litersPerSqmPerSecond;
          desc.hudOverlayFilename = hudOverlayFilename;
  
 
          AP900.sprayTypes[name] = desc;
          AP900.sprayTypeIndexToDesc[AP900.NUM_SPRAYTYPES] = desc;
  
          local fillType = Fillable.registerFillType(name)
          AP900.sprayTypeToFillType[AP900.NUM_SPRAYTYPES] = fillType;
          AP900.fillTypeToSprayType[fillType] = AP900.NUM_SPRAYTYPES;
      end;
end;
  
  AP900.registerSprayType("fertilizer", 0.3, 0.5, "");
  AP900.registerSprayType("manure", 0.01, 0.5, "");
  AP900.registerSprayType("liquidManure", 0.01, 0.5, "");
  
function AP900.prerequisitesPresent(specializations)
      return SpecializationUtil.hasSpecialization(Fillable, specializations) and SpecializationUtil.hasSpecialization(PowerShaft, specializations) and SpecializationUtil.hasSpecialization(Animation_AP900, specializations);
end;
 
function AP900:load(xmlFile)
  
      assert(self.setIsTurnedOn == nil, "AP900 needs to be the first specialization which implements setIsTurnedOn");
      self.setIsTurnedOn = AP900.setIsTurnedOn;
      self.getIsTurnedOnAllowed = AP900.getIsTurnedOnAllowed;
  
      assert(self.setIsSprayerFilling == nil, "AP900 needs to be the first specialization which implements setIsSprayerFilling");
      self.setIsSprayerFilling = AP900.setIsSprayerFilling;
      self.addSprayerFillTrigger = AP900.addSprayerFillTrigger;
      self.removeSprayerFillTrigger = AP900.removeSprayerFillTrigger;
  
      self.fillLitersPerSecond = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.fillLitersPerSecond"), 500);
      self.isSprayerFilling = false;
  
      self.sprayLitersPerSecond = {};
      local i=0;
      while true do
          local key = string.format("vehicle.sprayUsages.sprayUsage(%d)", i);
          if not hasXMLProperty(xmlFile, key) then
              break;
          end;
          local fillType = getXMLString(xmlFile, key.. "#fillType");
          local litersPerSecond = getXMLFloat(xmlFile, key.. "#litersPerSecond");
          if fillType ~= nil and litersPerSecond ~= nil then
              local fillTypeInt = Fillable.fillTypeNameToInt[fillType];
              if fillTypeInt ~= nil then
                  self.sprayLitersPerSecond[fillTypeInt] = litersPerSecond;
                  if self.defaultSprayLitersPerSecond == nil then
                      self.defaultSprayLitersPerSecond = litersPerSecond;
                  end;
              else
                  print("Warning: Invalid spray usage fill type '"..fillType.."' in '" .. self.configFileName.. "'");
              end;
          end;
          i = i+1;
      end;
      if self.defaultSprayLitersPerSecond == nil then
          print("Warning: No spray usage specified for '" .. self.configFileName.. "'. This sprayer will not use any spray.");
         self.defaultSprayLitersPerSecond = 0;
      end;
  
      self.sprayValves = {};
  
      if self.isClient then
          local psFile = getXMLString(xmlFile, "vehicle.sprayParticleSystem#file");
          if psFile ~= nil then
              local i=0;
              while true do
                  local baseName = string.format("vehicle.sprayValves.sprayValve(%d)", i);
                  local node = getXMLString(xmlFile, baseName.. "#index");
                  if node == nil then
                      break;
                  end;
                  node = Utils.indexToObject(self.components, node);
                  if node ~= nil then
                      local sprayValve = {};
                      sprayValve.particleSystems = {};
                      Utils.loadParticleSystem(xmlFile, sprayValve.particleSystems, "vehicle.sprayParticleSystem", node, false, nil, self.baseDirectory);
                      table.insert(self.sprayValves, sprayValve);
                  end;
                  i = i+1;
              end;
          end;
  
          local spraySound = getXMLString(xmlFile, "vehicle.spraySound#file");
          if spraySound ~= nil and spraySound ~= "" then
              spraySound = Utils.getFilename(spraySound, self.baseDirectory);
              self.spraySound = createSample("spraySound");
              self.spraySoundEnabled = false;
              loadSample(self.spraySound, spraySound, false);
              self.spraySoundPitchOffset = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.spraySound#pitchOffset"), 1);
              self.spraySoundVolume = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.spraySound#volume"), 1);
          end;
      end;
  
      self.sprayerFillTriggers = {};
      self.sprayerFillActivatable = SprayerFillActivatable:new(self);
  
      self.isTurnedOn = false;
      self.speedViolationMaxTime = 1000;
      self.speedViolationTimer = self.speedViolationMaxTime;
	  
	self.rotateSpinners = {};
	local i=0;
	while true do
		local baseName = string.format("vehicle.rotateSpinners.node(%d)", i);
		local index = getXMLString(xmlFile, baseName.. "#index");
		local x,y,z = Utils.getVectorFromString(getXMLString(xmlFile, baseName.. "#rotationSpeed"));
		local rotationSpeed = {x,y,z};
		local runOutTime = Utils.getNoNil(getXMLFloat(xmlFile, baseName.. "#runOutTime"), 2)*1000;
		if index == nil or rotationSpeed == nil or runOutTime == nil then
			break;
		end;
		local node = Utils.indexToObject(self.components, index);
		if node ~= nil then
			local entry = {};
			entry.node = node;
			entry.runOutTime = runOutTime;
			entry.rotationSpeedMax = rotationSpeed;
			entry.rotationSpeedMin = {0,0,0};
			entry.rotationSpeedCurrent = {0,0,0};
			table.insert(self.rotateSpinners, entry);
		end;
		i = i+1;
	end;	
	
	local rotarySound = getXMLString(xmlFile, "vehicle.rotarySound#file");
	if rotarySound ~= nil and rotarySound ~= "" then
		rotarySound = Utils.getFilename(rotarySound, self.baseDirectory);
		self.rotarySound = createSample("rotarySound");
		loadSample(self.rotarySound, rotarySound, false);
		self.rotarySoundPitchOffset = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.rotarySound#pitchOffset"), 0);
		self.rotarySoundVolume = 0.0;
		self.rotarySoundVolumeMin = 0.0;
		self.rotarySoundVolumeMax = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.rotarySound#volume"), 1.0);
		self.rotarySoundEnabled = false;
	end;

	  self.spreaderParticleSystems = {};
      local i = 0;
      while true do
        local namei = string.format("vehicle.spreaderParticleSystems.spreaderParticleSystems(%d)", i);
		local nodei = Utils.indexToObject(self.components, getXMLString(xmlFile, namei .. "#index"));
		if nodei == nil then
			break;
		end; 
        Utils.loadParticleSystem(xmlFile, self.spreaderParticleSystems, namei, nodei, false, nil, self.baseDirectory)		
		i = i +1;		
      end;
	  
 	  self.setVehicleRpmUp = SpecializationUtil.callSpecializationsFunction("setVehicleRpmUp");
	  self.saveMinRpm = 0; 
      self.printWarningTime = 0;
end;
  
function AP900:delete()
  
      g_currentMission:removeActivatableObject(self.sprayerFillActivatable);
  
      for k,sprayValve in pairs(self.sprayValves) do
          Utils.deleteParticleSystem(sprayValve.particleSystems);
      end;
  
      if self.spraySound ~= nil then
          delete(self.spraySound);
      end;
	if self.rotarySound ~= nil then
		delete(self.rotarySound);
	end;
 	Utils.deleteParticleSystem(self.spreaderParticleSystems);
end;
  
function AP900:readStream(streamId, connection)
      local turnedOn = streamReadBool(streamId);
      local isSprayerFilling = streamReadBool(streamId);
      self:setIsTurnedOn(turnedOn, true);
      self:setIsSprayerFilling(isSprayerFilling, true);
end;
  
function AP900:writeStream(streamId, connection)
      streamWriteBool(streamId, self.isTurnedOn);
      streamWriteBool(streamId, self.isSprayerFilling);
end;
  
function AP900:readUpdateStream(streamId, timestamp, connection)
end;
  
function AP900:writeUpdateStream(streamId, connection, dirtyMask)
end;
  
function AP900:mouseEvent(posX, posY, isDown, isUp, button)
end;
  
function AP900:keyEvent(unicode, sym, modifier, isDown)
end;
 
function AP900:update(dt)
  
      if self.isClient then
          if self:getIsActiveForInput() then
             if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA) then
                  self:setIsTurnedOn(not self.isTurnedOn);
             end;
			if InputBinding.isPressed(InputBinding.IMPLEMENT_EXTRA) and self.PTOId then
				self.printWarningTime = self.time + 1500;
			end;
		end;
      end;
	
	for i, jointDesc in pairs(self.componentJoints) do
		setJointFrame(self.componentJoints[i].jointIndex, 0, self.componentJoints[i].jointNode);
	end;	 
end;
  
function AP900:updateTick(dt)

		if self:getIsActive() then
			if not self.attacherVehicle.isMotorStarted then
				self.isTurnedOn = false;
			end;
			if self.PTOId then
				self.isTurnedOn = false;
			end;
			local isKeyEvent = false;	
			if self.isTurnedOn then
				isKeyEvent = true;	
			else
				isKeyEvent = false;	
			end;
			self:setVehicleRpmUp(dt, isKeyEvent);	
			local volume = Utils.getMovedLimitedValues({self.rotarySoundVolume}, {self.rotarySoundVolumeMax}, {self.rotarySoundVolumeMin}, 1, 1200, dt, not self.isTurnedOn);
			self.rotarySoundVolume = volume[1];
				if self.isTurnedOn and not self.PTOId then
					if not self.rotarySoundEnabled and self:getIsActiveForSound() then
						playSample(self.rotarySound, 0, self.rotarySoundVolume, 0);
						setSamplePitch(self.rotarySound, self.rotarySoundPitchOffset);
						self.rotarySoundEnabled = true;
					end;
					if self.rotarySoundVolume < self.rotarySoundVolumeMax then
						setSampleVolume(self.rotarySound, self.rotarySoundVolume);
					end;
					else
					if self.rotarySoundEnabled then
						if self.rotarySoundVolume <= self.rotarySoundVolumeMin then
							stopSample(self.rotarySound);
							self.rotarySoundEnabled = false;
						else
							setSampleVolume(self.rotarySound, self.rotarySoundVolume);
						end;
					end;

				end;	
        if self.isTurnedOn then
			if self.animationParts[1].currentPosition > 200 then 
				  self.lastSprayingArea = 0;
	  
				  if self:doCheckSpeedLimit() and self.lastSpeed*3600 > 31 then
					  self.speedViolationTimer = self.speedViolationTimer - dt;
				  else
					  self.speedViolationTimer = self.speedViolationMaxTime;
				  end;
	  
				  if self.isServer then
					  if self.speedViolationTimer > 0 then
							local litersPerSecond = self.sprayLitersPerSecond[self.currentFillType];
							if self.animationParts[1].currentPosition > 200 and self.animationParts[1].currentPosition <= 1000 then 
								litersPerSecond = litersPerSecond/9;
							elseif self.animationParts[1].currentPosition >= 1001 and self.animationParts[1].currentPosition <= 2000 then 
								litersPerSecond = litersPerSecond/6;
							elseif self.animationParts[1].currentPosition >= 2001 and self.animationParts[1].currentPosition <= 3000 then 
								litersPerSecond = litersPerSecond/3;					
							elseif self.animationParts[1].currentPosition >= 3001 and self.animationParts[1].currentPosition <= 4000 then 
								litersPerSecond = litersPerSecond;										
							end;	
								
						  if litersPerSecond == nil then
							  litersPerSecond = self.defaultSprayLitersPerSecond;
						  end
						  local usage = litersPerSecond * dt*0.001;
	  
						  local hasSpray = false;
	  
						  if self.capacity == 0 or self:getIsHired() then
							  hasSpray = true;
							  local sprayType = AP900.fillTypeToSprayType[self.currentFillType];
							  if sprayType ~= nil then
								  local sprayTypeDesc = AP900.sprayTypeIndexToDesc[sprayType];
								  local delta = usage*sprayTypeDesc.pricePerLiter
								  g_currentMission.missionStats.expensesTotal = g_currentMission.missionStats.expensesTotal + delta;
								  g_currentMission.missionStats.expensesSession = g_currentMission.missionStats.expensesSession + delta;
	  
								  g_currentMission:addSharedMoney(-delta);
	  
								  -- TODO update spray stats
								  --[[g_currentMission.missionStats.sprayUsageTotal = g_currentMission.missionStats.sprayUsageTotal + usage;
								  g_currentMission.missionStats.sprayUsageSession = g_currentMission.missionStats.sprayUsageSession + usage;]]
							  end;
						  else
							  if self.fillLevel > 0 then
								  hasSpray = true;
								  self:setFillLevel(self.fillLevel - usage, self.currentFillType);
							  end;
						  end;
	  
						  if hasSpray then
							  local cuttingAreasSend = {};
							  for k,cuttingArea in pairs(self.cuttingAreas) do
								  if self:getIsAreaActive(cuttingArea) then
									  local x,y,z = getWorldTranslation(cuttingArea.start);
									  local x1,y1,z1 = getWorldTranslation(cuttingArea.width);
									  local x2,y2,z2 = getWorldTranslation(cuttingArea.height);
	  
									  local sqm = math.abs((z1-z)*(x2-x) - (x1-x)*(z2-z)); -- this is the cross product with y=0
	  
									  --Utils.updateSprayArea(x, z, x1, z1, x2, z2);
	  
									  self.lastSprayingArea = self.lastSprayingArea + sqm;
	  
									  table.insert(cuttingAreasSend, {x,z,x1,z1,x2,z2});
								  end;
							  end;
							  if (table.getn(cuttingAreasSend) > 0) then
								  SprayerAreaEvent.runLocally(cuttingAreasSend);
								  g_server:broadcastEvent(SprayerAreaEvent:new(cuttingAreasSend));
							  end;
						  end;
					  end;
				  end;
	  
				  if self.isClient then
					  if not self.spraySoundEnabled and self:getIsActiveForSound() then
						  playSample(self.spraySound, 0, self.spraySoundVolume, 0);
						  setSamplePitch(self.spraySound, self.spraySoundPitchOffset);
						  self.spraySoundEnabled = true;
					  end;
				  end;
	  
				  if self.fillLevel <= 0 and self.capacity ~= 0 then
					  self:setIsTurnedOn(false, true);
				  end;
			  else
				  self.speedViolationTimer = self.speedViolationMaxTime;
			  end;
		 end;
     end;
  
      if self.isSprayerFilling and self.isServer then
          local disableFilling = false;
          if self:allowFillType(self.sprayerFillingFillType, false) then
              local oldFillLevel = self.fillLevel;
  
              local delta = self.fillLitersPerSecond*dt*0.001;
  
              local silo = g_currentMission:getSiloAmount(self.sprayerFillingFillType);
              if self.sprayerFillingIsSiloTrigger then
                  if silo <= 0 then
                     disableFilling = true;
                  end;
                  delta = math.min(delta, silo);
              end;
  
              self:setFillLevel(self.fillLevel + delta, self.sprayerFillingFillType, true);
              local delta = self.fillLevel - oldFillLevel;
              if delta > 0 then
                  if self.sprayerFillingIsSiloTrigger then
                      g_currentMission:setSiloAmount(self.sprayerFillingFillType, silo - delta);
                  else
                      local sprayType = AP900.fillTypeToSprayType[self.sprayerFillingFillType];
                      if sprayType ~= nil then
                          local sprayTypeDesc = AP900.sprayTypeIndexToDesc[sprayType]
  
                          local price = delta*sprayTypeDesc.pricePerLiter;
                          g_currentMission.missionStats.expensesTotal = g_currentMission.missionStats.expensesTotal + price;
                          g_currentMission.missionStats.expensesSession = g_currentMission.missionStats.expensesSession + price;
  
                          g_currentMission:addSharedMoney(-price);
                      end;
                  end;
              elseif self.fillLevel == self.capacity then
                  disableFilling = true;
              end;
          else
              disableFilling = true;
          end;
          if disableFilling then
              self:setIsSprayerFilling(false);
          end;
      end;
  		for k, spinner in pairs(self.rotateSpinners) do
			local values = Utils.getMovedLimitedValues(spinner.rotationSpeedCurrent, spinner.rotationSpeedMax, spinner.rotationSpeedMin, 3, spinner.runOutTime, dt, not self.isTurnedOn);
			spinner.rotationSpeedCurrent = values;
			rotate(spinner.node, unpack(spinner.rotationSpeedCurrent));
		end; 
		if self.isTurnedOn and not self.PTOId and self.animationParts[1].currentPosition > 200 then
			Utils.setEmittingState(self.spreaderParticleSystems, true);
		else
			Utils.setEmittingState(self.spreaderParticleSystems, false);
		end;
end;
  
function AP900:draw()
  
      if self.isClient then
          if self.fillLevel <= 0 and self.capacity ~= 0 then
              g_currentMission:addExtraPrintText(g_i18n:getText("FirstFillTheTool"));
          end;
  
          if self.isTurnedOn then
              g_currentMission:addHelpButtonText(string.format(g_i18n:getText("turn_off_OBJECT"), self.typeDesc), InputBinding.IMPLEMENT_EXTRA);
          else
              g_currentMission:addHelpButtonText(string.format(g_i18n:getText("turn_on_OBJECT"), self.typeDesc), InputBinding.IMPLEMENT_EXTRA);
          end;
  
          if math.abs(self.speedViolationTimer - self.speedViolationMaxTime) > 2 then
              g_currentMission:addWarning(g_i18n:getText("Dont_drive_to_fast") .. "\n" .. string.format(g_i18n:getText("Cruise_control_levelN"), "2", InputBinding.getKeyNamesOfDigitalAction(InputBinding.SPEED_LEVEL2)), 0.07+0.022, 0.019+0.029);
          end;
		if self.printWarningTime > self.time then
			g_currentMission:addWarning(g_i18n:getText("turnON_Error"), 0.018, 0.033);
		end;
      end;
end;

function AP900:onAttach(attacherVehicle)
	self.attacherVehicle = attacherVehicle;
	if self.attacherVehicleCopy == nil then
		self.attacherVehicleCopy = self.attacherVehicle;
	end;
	self.saveMinRpm = self.attacherVehicle.motor.minRpm;
end;
  
function AP900:onDetach()
      if self.deactivateOnDetach then
          AP900.onDeactivate(self);
      else
         AP900.onDeactivateSounds(self);
      end;
	  
	if self.spreaderParticleSystems ~= nil then
		Utils.setEmittingState(self.spreaderParticleSystems, false);
	end; 
	for k, steerable in pairs(g_currentMission.steerables) do
		if self.attacherVehicleCopy == steerable then
			steerable.motor.minRpm = self.saveMinRpm;
			self.attacherVehicleCopy = nil;
		end;
	end;
end;
  
function AP900:onLeave()
      if self.deactivateOnLeave then
          AP900.onDeactivate(self);
      else
          AP900.onDeactivateSounds(self);
      end;
	if self.spreaderParticleSystems ~= nil then
		Utils.setEmittingState(self.spreaderParticleSystems, false);
	end; 
end;
  
function AP900:onDeactivate()
    self.speedViolationTimer = self.speedViolationMaxTime;
    self:setIsTurnedOn(false, true)
    AP900.onDeactivateSounds(self);
	if self.spreaderParticleSystems ~= nil then
		Utils.setEmittingState(self.spreaderParticleSystems, false);
	end; 
end;
  
function AP900:onDeactivateSounds()
    if self.spraySoundEnabled then
        stopSample(self.spraySound);
        self.spraySoundEnabled = false;
    end;
	if self.rotarySoundEnabled then
		stopSample(self.rotarySound);
		self.rotarySoundEnabled = false;
	end;
end;
  
function AP900:getIsTurnedOnAllowed(isTurnedOn)
      if not isTurnedOn or self.fillLevel > 0 or self.capacity == 0 then
          return true;
      end;
end;
  
function AP900:setIsTurnedOn(isTurnedOn, noEventSend)
      if isTurnedOn ~= self.isTurnedOn then
          if self:getIsTurnedOnAllowed(isTurnedOn) then
              SetTurnedOnEvent.sendEvent(self, isTurnedOn, noEventSend)
  
              self.isTurnedOn = isTurnedOn;
              if self.isClient then
                  for k,sprayValve in pairs(self.sprayValves) do
                      Utils.setEmittingState(sprayValve.particleSystems, self.isTurnedOn);
                  end;
  
                  if not self.isTurnedOn and self.spraySoundEnabled then
                      stopSample(self.spraySound);
                      self.spraySoundEnabled = false;
                  end;
              end;
              self.speedViolationTimer = self.speedViolationMaxTime;
          end;
      end;
end;

function AP900:setVehicleRpmUp(dt, isActive)
	if self.attacherVehicle ~= nil and self.saveMinRpm ~= 0 then
		if dt ~= nil then
			if isActive == true then
				self.attacherVehicle.motor.minRpm = math.max(self.attacherVehicle.motor.minRpm-dt, -1000);
			else
				self.attacherVehicle.motor.minRpm = math.min(self.attacherVehicle.motor.minRpm+dt*2, self.saveMinRpm);
			end;
		else
			self.attacherVehicle.motor.minRpm = self.saveMinRpm;
		end;
		if self.attacherVehicle.isMotorStarted then
			local fuelUsed = 0.0000011*math.abs(self.attacherVehicle.motor.minRpm);
			self.attacherVehicle:setFuelFillLevel(self.attacherVehicle.fuelFillLevel-fuelUsed);
			g_currentMission.missionStats.fuelUsageTotal = g_currentMission.missionStats.fuelUsageTotal + fuelUsed;
			g_currentMission.missionStats.fuelUsageSession = g_currentMission.missionStats.fuelUsageSession + fuelUsed;
		end;
	end;
end;
  
function AP900:setIsSprayerFilling(isFilling, fillType, isSiloTrigger, noEventSend)
      SprayerSetIsFillingEvent.sendEvent(self, isFilling, fillType, isSiloTrigger, noEventSend)
      if self.isSprayerFilling ~= isFilling then
          self.isSprayerFilling = isFilling;
          self.sprayerFillingFillType = fillType;
          self.sprayerFillingIsSiloTrigger = isSiloTrigger;
      end;
end;
  
function AP900:addSprayerFillTrigger(trigger)
      if table.getn(self.sprayerFillTriggers) == 0 then
          g_currentMission:addActivatableObject(self.sprayerFillActivatable);
      end;
      table.insert(self.sprayerFillTriggers, trigger);
end;
  
function AP900:removeSprayerFillTrigger(trigger)
      for i=1, table.getn(self.sprayerFillTriggers) do
          if self.sprayerFillTriggers[i] == trigger then
              table.remove(self.sprayerFillTriggers, i);
              break;
          end;
      end;
      if table.getn(self.sprayerFillTriggers) == 0 then
          if self.isServer then
              self:setIsSprayerFilling(false);
          end;
          g_currentMission:removeActivatableObject(self.sprayerFillActivatable);
      end;
end;
  
  SprayerFillActivatable = {}
  local SprayerFillActivatable_mt = Class(SprayerFillActivatable);
  
function SprayerFillActivatable:new(sprayer)
      local self = {};
      setmetatable(self, SprayerFillActivatable_mt);
  
      self.sprayer = sprayer;
      self.activateText = "unknown";
  
      self.currentTrigger = nil;
  
      return self;
end;
  
  
function SprayerFillActivatable:getIsActivatable()
      self.currentTrigger = nil;
      if not self.sprayer:getIsActiveForInput() or self.sprayer.fillLevel == self.sprayer.capacity then
          return false;
      end;
      -- find the first trigger which is activable
      for i=1, table.getn(self.sprayer.sprayerFillTriggers) do
          local trigger = self.sprayer.sprayerFillTriggers[i];
          if trigger:getIsActivatable(self.sprayer) then
              self.currentTrigger = trigger;
              self:updateActivateText();
              return true;
          end;
      end;
      return false;
end;
  
function SprayerFillActivatable:onActivateObject()
      self.sprayer:setIsSprayerFilling(not self.sprayer.isSprayerFilling, self.currentTrigger.fillType, self.currentTrigger.isSiloTrigger);
      self:updateActivateText();
      g_currentMission:addActivatableObject(self);
end;
  
function SprayerFillActivatable:drawActivate()
      -- TODO draw icon
end;
  
function SprayerFillActivatable:updateActivateText()
      if self.sprayer.isSprayerFilling then
          self.activateText = string.format(g_i18n:getText("stop_refill_OBJECT"), self.sprayer.typeDesc);
      else
          self.activateText = string.format(g_i18n:getText("refill_OBJECT"), self.sprayer.typeDesc);
     end;
end;